

#######################################################################
# Copyright (c) 2019 Dylan Hinson, Josh Luthy and Ian Taylor of BD Life
# Sciences - Informatics, Ashland, OR
#
# R Script Template Designed and Written by: Josh Luthy and Ian Taylor
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################


# This is a Shiny web application for interacting with visualizations from FlowJo and SeqGeq. 
# You can run the application by clicking the 'Run App' button above within RStudio, or by launching 
#   the "interactive" portion of the ViolinBox plugin.

# packages<-function(x){
#   x<-as.character(match.call()[[2]])
#   if(!require(x,character.only=TRUE)){
#     install.packages(pkgs=x,repos="http://cran.r-project.org")
#     require(x,character.only=TRUE)
#   }
# }
#
# packages(shiny)
# packages(shinydashboard)
# packages(readr)
# packages(dplyr)
# packages(pheatmap)
# packages(viridisLite)
# packages(DT)
# packages(ggplot2)
# packages(cowplot)
# packages(shinyWidgets)
# packages(ggsignif)
# packages(broom)
# packages(ggpubr)
# packages(gginnards)
# packages(tidyverse)
# packages(gtools)
# packages(sunburstR)
# packages(plotly)
# packages(RColorBrewer)

#######################
## Global Settings:  ##
#######################
# inCSVPath <- "/Users/10201397/Desktop/Shiny_ViolinBox/extData/B1.smallPop.catPar.ExtNode.csv"
inCSVPath <- "FJ_DATA_FILE_PATH"
# catParRaw <- "FJ_CAT_PAR"

outPath <- "FJ_OUTS_FOLDER"
popMap <- "FJ_CATPOP_MAP"

popLoc <- paste0(outPath,"/popH.txt")

popMap <- readr::read_file(popLoc)

# if(catParRaw != "{}" && catParRaw != ""){
popMap <- gsub(", ", "C=O=D=E=X", popMap)
popMap <- gsub(" ", "_", popMap)
popMap <- gsub("C=O=D=E=X", ", ", popMap)
popMap <- gsub("\\.", "_", popMap)
# popMap <- gsub("\\=", "Eq", popMap)
# }

# df <- read.csv(inCSVPath, sep = ",", header = TRUE, check.names = FALSE)

# if(!nrow(df) > 0){
#   print("Unable to read the input file correctly, try again.")
#   stopAndReturnError("Unable to read the input file correctly.", call. = FALSE)
# }

## We need to make sure that there are or are not categorical parameters
# noKatPar <- TRUE
# if(catParRaw != "[]" && catParRaw != ""){
#   catParam <- gsub("(\\[)","",gsub("(\\])","",catParRaw))
#   catPar <- as.symbol(catParam)
#   noKatPar <- FALSE
# }

## Detectedr if this is FJ or SG
# if ("EventNumberDP" %in% colnames(df)){
#   print("EventNumber Detected - This is FlowJo")
#   df <- df %>%
#     select(EventNumberDP, everything())
#   colnames(df) <- gsub( "(^.*)( :: )" , "", colnames(df), ignore.case = TRUE )
#   isSeqGeq <- F
#   df$EventNumberDP <- NULL
# } else {
#   print("No EventNumber Detected - We're in SeqGeq")
#   isSeqGeq <- T
# }


# if (!noKatPar){

## Remove the "." dots from categorical parameters too...
# catParamTemp <- gsub("\\.","_",catParam)
# catParamTemp <- gsub("\\s","_",catParamTemp)
# catParamTemp <- gsub("\\=", "Eq", catParamTemp)
# names(df)[names(df) == catParam] <- catParamTemp
# catParam <- catParamTemp

## SET UP the POP NAMES
hasPopMap <- F
firstRound <- T
if(popMap != "" && popMap != "{}" && popMap != "[]"){
    hasPopMap <- T

    popMapNew <- gsub("\\}","",gsub("\\{","",gsub("\\s","",popMap)))
    splitMap <- as.data.frame(strsplit(popMapNew,","))

    dataMap <- data.frame( popNumber = as.character(),
    ValueNumber = character(),
    ParentName = character(),
    CountNumber = character(),
    Name = character(),
    stringsAsFactors = F )

    for(x in 1:nrow(splitMap)){

        xthSet <- as.data.frame(strsplit(as.character(splitMap[x,]),"<=>"))

        wthSet <- as.data.frame(strsplit(as.character(xthSet[1,]),"THISEQUALSSIGN"))

        ythSet <- as.data.frame(strsplit(as.character(xthSet[2,]),"[=]"))
        zthSet <- as.data.frame(strsplit(as.character(wthSet[1,]),"="))

        checkMe <- as.numeric(gsub("\\[","",ythSet[1,]))

        if(checkMe > 0){
            dataMap[x,] <- data.frame(popNumber = as.character(as.numeric(zthSet[1,])),
            ValueNumber = as.character(zthSet[2,]), #as.character(zthSet[1,]), ## sig figs set to three to match what java is doing. NOTE: This is all for the sake of FlowSOM
            ParentName = as.character(wthSet[2,]),
            CountNumber = as.character(gsub("\\[","",ythSet[1,])),
            Name = as.character(gsub("\\]","",ythSet[2,])),
            stringsAsFactors = F)
        }
    }
    dataMap <- na.omit(dataMap)
}

foo <- dataMap # This is just a save game
# dataMap <- foo

### This section is super important - It makes sure that the children have <= the number of cells of the parent
## Initialize some values for the first round of this loop:
prePreviousParent <- ""
CountValueOld <- 0
mthParent <- dataMap[1,2]
q = nrow(dataMap)
for (m in 1:nrow(dataMap)){

    CountValueNow <- as.numeric(dataMap[m,4]) ## Get the count of the first loop
    previousParent <- mthParent ## Set the previous parent for checking if it changes later
    mthParent <- dataMap[m,3] ## Set the parent current parent value

    if( (previousParent == mthParent) && (mthParent != "") && (m != q) ){ ## If the parent value hasn't changed add the count of this iteration to the current set of counts ("countOld"), making sure not to use the root. Also,
        # if this is the last iteration we need to switch the statement here!
        CountValueOld <- CountValueNow + CountValueOld
    } else if(mthParent != "") { ## If the previous parent is not the same as the current parent and not root then...
        prePreviousParent <- previousParent ## ... Set that previous parent as the "prePreviousParent" in case we need that later

        for (r in 1:nrow(dataMap)){ ## ... get the rowname for the previous parent
            if(previousParent == "") {
                rowOfMthParent <- row.names(dataMap[1,])
            }else if(dataMap[r,5] == previousParent){
                rowOfMthParent <- row.names(dataMap[r,])
            }
        }

        if( m == q && (previousParent == mthParent) ) {
            CountValueOld <- (CountValueOld + CountValueNow)
        }

        as.numeric(dataMap[rowOfMthParent,4])

        if(CountValueOld > as.numeric(dataMap[rowOfMthParent,4]) ){
            dataMap[rowOfMthParent,4] <- as.character(CountValueOld) ## ... if the counted values are larger than the count set in our data matrix set the new count as our parent's count
        }
        CountValueOld <- CountValueNow ## ... set the current count in the slot for continueing the counting going forward
    } else {
        writeLines("Sunburst Data Frame Created")
    }
}


#######################
## CREATE FUNCTIONS  ##
#######################

## NA ##

## Input CSV for testing purposes ##
setwd(outPath)


#####################################################
##  Converting FlowJo Outputs into Sunburst Format ##
#####################################################

sp <- length(dataMap)

sun_labels = (dataMap$Name)
sun_parents = gsub("SPACEDASHSPACE", " - ", gsub("PARENTNODEPARENT","",dataMap$ParentName))
sun_values = (dataMap$CountNumber)

sunny_ids = gsub("SPACEDASHSPACE"," - ",dataMap$ValueNumber)

getPalette <- colorRampPalette(brewer.pal(8, 'Set1'))(sp)
d <- data.frame(sun_parents, sun_labels, sun_values, sunny_ids, stringsAsFactors = FALSE)


####################
# Set Menu Options #
####################

# sbPalette <- c("magma", "plasma", "inferno", "viridis", "cividis") ##, "rainbow")
sbPalette <- c("Dark2","Accent","Pastel2","Set2", "Pastel1", "Set1","Paired","Set3","Spectral","BrBG","PiYG","PRGn","PuOr")
orientationOptions <- c("horizontal", "radial", "tangential", "auto")
sliceNorm <- c('remainder', 'total', "relative")
# correction <- c("bonferroni", "fdr")
# plotTypes <- c("violin","jitter","box")


#######################
##   Shiny Part:     ##
#######################

uClust <- dataMap$Name

Sunbursting <- function() {
    require(shiny)
    # shinyApp(ui = ui, server = server)
    library(shiny)
    library(sunburstR)
    library(plotly)
    options(shiny.trace=TRUE)

    shinyApp(
    ui <- dashboardPage(
    dashboardHeader(title = "Sunburst"),
    mysidebar <- dashboardSidebar(
    width = 190,
    sidebarMenu(
    id = 'tabs',

    actionButton("refreshPlot",
    label = "RECALCULATE"),

    menuItem(text = "Sunburst",
    tabName = "sunburstButton",
    startExpanded = F,
    icon = icon("superpowers"),

    menuSubItem("", tabName ="sunny_plotter", icon = NULL),

    selectInput("plotColor",
    label = "Select Colors",
    choices = sbPalette,
    multiple = FALSE,
    selectize = TRUE),

    selectInput("textOrientation",
    label = "Text Orientation",
    choices = orientationOptions,
    multiple = FALSE,
    selected = "auto"),

    selectInput("normy",
    label = "Slice Normalization",
    choices = sliceNorm,
    multiple = F,
    selected = "remainder"),

    numericInput("maxDepth",
    label = "Maximum Layers",
    value = NULL,
    max = 20,
    min = 2,
    step = 1
    ),

    downloadButton("print", label = "PRINT HTML"),
    actionButton("show", "Help")

    )
    )
    ),

    dashboardBody(

    ## These CSS sytles will giv us the full viewport height for graphs specified:
    tags$head(tags$style("#heatmapButton{height:90vh !important;}")),
    tags$head(tags$style("#violins{height:90vh !important;}")),

    tabItems(
    # tabItem(tabName = "data",
    #         DTOutput(outputId = "table1")
    # ),

    tabItem(tabName = "sunny_plotter",
    plotlyOutput("sunny_plotter")
    )
    # tabItem(tabName = "print",
    # plotlyOutput("print")
    # )
    ),


    tags$head(
    tags$link(rel="icon", href="http://fjinstallers.s3-us-west-2.amazonaws.com/Plugins/sunburstIcon.png"),
    tags$style(HTML('

        /* logo */
        .skin-blue .main-header .logo {
                              background-color: #696969;
        }

        #                       /* logo when hovered */
        # .skin-blue .main-header .logo:hover {
        #                       background-color: ##696969;
        #                       }

        /* navbar (rest of the header) */
        .skin-blue .main-header .navbar {
                              background-color: #696969;
                              }

        /* main sidebar */
        .skin-blue .main-sidebar {
                              background-color: #696969;
                              }

        /* active selected tab in the sidebarmenu */
        .skin-blue .main-sidebar .sidebar .sidebar-menu .active a{
                              background-color: #b6a4b6;
                              }

        /* other links in the sidebarmenu */
        .skin-blue .main-sidebar .sidebar .sidebar-menu a{
                              background-color: #b0b0b0;
                              color: #000000;
                              }

        /* other links in the sidebarmenu when hovered */
         .skin-blue .main-sidebar .sidebar .sidebar-menu a:hover{
                              background-color: #ff69b4;
                              }
        /* toggle button when hovered  */
         .skin-blue .main-header .navbar .sidebar-toggle:hover{
                              background-color: #ff69b4;
                              }
        /* body */
          .content-wrapper, .right-side {
                              background-color: #DCDCDC;
                              }
      ')))

    )
    ),


    #################
    # Make a Server #
    #################

    server <- function(input, output, session) {

        session_store <- reactiveValues()

        observeEvent(input$sidebarItemExpanded, {
            test <- input$sidebarItemExpanded
            if(test == "sunny_plotter"){
                updateTabItems(session,"tabs","sunny_plotter")
            }
        })
        observeEvent(input$show, {
            showNotification("To Save as HTML First Download Pandoc Application Here: ",
            a(href="https://pandoc.org/installing.html", "pandoc.org")) ## target="_blank">pancod.org</a>)
        })

        getPalette <- eventReactive(input$refreshPlot, {
            numUniqueParams <- 0
            # numUniqueParams <- length(d$sun_labels)
            if (numUniqueParams < 6) {
                numUniqueParams <- 6
            }
            yourPal <- colorRampPalette(brewer.pal(8, input$plotColor))(numUniqueParams)
            return(yourPal)
        })

        ###############################
        ##      Make sunburst plot  ##
        ###############################
        sunburstPlot <- eventReactive(input$refreshPlot, {
            # TEST <<- d
            SB <- plot_ly(
            data = d,
            ids = ~sunny_ids,
            labels = ~sun_labels,         # The labels indicate the name of the population at the given hierarchy.
            parents = ~sun_parents,       # The parents indicate which place in the hierarchy the above label belongs to.
            values = ~sun_values,         # The values indicate what fraction of the parent the label gets.
            branchvalues = input$normy,
            insidetextorientation= input$textOrientation,
            # marker = list(color = "Total_NK"),
            maxdepth = input$maxDepth,
            type = 'sunburst'
            )
            colorTheSun <- getPalette()
            layout(SB, colorway = colorTheSun)
            # return(SB)
        })

        output$sunny_plotter <- renderPlotly({
            session_store$plt <- sunburstPlot()
            session_store$plt
        })

        output$print <- downloadHandler(
        filename = function() {
            paste("Sunburst.html", sep = "")
            # paste("File.html", sep = "")
        },
        content = function(file) {
            saveWidget(as_widget(session_store$plt), file, selfcontained = T)
        }
        )

        launch.browser=TRUE
        session$onSessionEnded(stopApp)
    })
}
Sunbursting()
