#######################################################################
# Copyright (c) 2019 Josh Luthy BD Life Sciences - Informatics,
# Ashland, OR
#
# R Script Template Designed and Written by: Josh Luthy and Ian Taylor
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################
 ## TRY TO AUTO INSTALL REQUIRED PACKAGES ##
 packages<-function(x){
   x<-as.character(match.call()[[2]])
   if(!require(x,character.only=TRUE)){
       install.packages(pkgs=x,repos="http://cran.r-project.org")
       require(x,character.only=TRUE)
   }
 }
## the Real Deal (aka required packages) ##
packages(tidyr)
packages(readr)
packages(RColorBrewer)
packages(ggplot2)
packages(plyr)
packages(dplyr)
packages(cowplot)
packages(viridis)
packages(pheatmap)
packages(Cairo)
packages(xml2)
packages(ggpubr)
packages(ggpval)
packages(ggsignif)
packages(xml2)
packages(shinydashboard)
packages(shiny)

#######################
# Required in order to load the geom_flat_violin function as it does not currently exist in a package
# source("https://gist.githubusercontent.com/benmarwick/2a1bb0133ff568cbe28d/raw/fb53bd97121f7f9ce947837ef1a4c65a73bffb3f/geom_flat_violin.R")
sessionInfo()

# cat("FJ-DATA-FILE-PATH         = FJ_DATA_FILE_PATH \n")
# cat("FJ-CSV-OUPUT-FILE         = FJ_CSV_OUPUT_FILE \n")
# cat("FJ-SAMPLE-NAME            = FJ_SAMPLE_NAME \n")
# cat("FJ-APPLY-ON               = FJ_APPLY_ON \n")
# cat("FJ-FLIP-GRAPH             = FJ_FLIP_GRAPH \n")
# cat("FJ-TRIM                   = FJ_TRIM \n")
# cat("FJ-JITTER                 = FJ_JITTER \n")
# cat("FJ-PDF                    = FJ_PDF \n")
# cat("FJ-REMOVE-BG              = FJ_REMOVE_BG \n")
# cat("FJ-HEIGHT                 = FJ_HEIGHT \n")
# cat("FJ-WIDTH                  = FJ_WIDTH \n")
# cat("FJ-WIDTH2                 = FJ_WIDTH2 \n")
# cat("FJ-LEGEND-SIZE            = FJ_LEGEND_SIZE \n")
# cat("FJ-JIT-SIZE               = FJ_JIT_SIZE \n")
# cat("FJ-JIT-DISP               = FJ_JIT_DISP \n")
# cat("FJ-JIT-DODGE              = FJ_JIT_DODGE \n")
# cat("FJ-MAX-SAMPLE-NAME-LEN    = FJ_MAX_SAMPLE_NAME_LEN \n")
# cat("FJ-LEGEND-POS             = FJ_LEGEND_POS \n")

##         LOAD INPUTS       ##
###############################

outPath <- "FJ_CSV_OUPUT_FILE"
inCSVPath <- "FJ_DATA_FILE_PATH"
sampleName <- "FJ_SAMPLE_NAME"
applyOn <- "FJ_APPLY_ON"
selFile <- "FJ_SELECTED_FILE"
extNodeEl <- "FJ_EXT_NODE_EL"
outs <- "FJ_OUTS_FOLDER"

parOverlayPlots <- FJ_OVERLAY_PLOTS
parFlipGraph <- FJ_FLIP_GRAPH
parTrim <- FJ_TRIM
parJitter <- FJ_JITTER
parViolin <- FJ_VIOLIN
parBox <- FJ_BOX
parCreatePDF <- FJ_PDF
parDarkTheme <- FJ_DARK_THEME

createStatTable <- FJ_STATS
parHeatmap <- FJ_HEAT_MAP
parScaledHM <- FJ_SCALEDHM
xmlPath <- "FJ_XML_PATH"

groupByClust <- FJ_GROUP_BY_CLUST
statCompare <- FJ_STAT_COMP

catParRaw <- "FJ_CAT_PAR"

noKatPar <- TRUE
 if(catParRaw != "[]" && catParRaw != ""){
    catParam <- gsub("(\\[)","",gsub("(\\])","",catParRaw))
    catPar <- as.symbol(catParam)
    noKatPar <- FALSE
 }


if(parDarkTheme){
    jColor <- "WHITE"
    fColor <- "BLACK"
} else {
    jColor <- "BLACK"
    fColor <- "WHITE"
}


## VESTIGIAL PARTS ##
#####################
# legendSize <- as.integer((parLegendSize * 24) / 100)
# jitterSize <- (parJitterSize * 0.5 / 100)
# jitterDisp <- parJitterDisp / 100
# jitterDodge <- parJitterDodge / 100
# if (parLegendPosition != "left" && parLegendPosition != "right") {
	## No extra width for legend if legend unless legend is on the left or right
#	parExtraWidthForLegend  <- 0
# }


##      CREATE FUNCTIONS     ##
###############################
## Half Violin Plot: ##
#######################

"%||%" <- function(a, b) {
  if (!is.null(a))
    a
  else
    b
}

geom_flat_violin <-
  function(mapping = NULL,
           data = NULL,
           stat = "ydensity",
           position = "dodge",
           trim = TRUE,
           scale = "area",
           show.legend = NA,
           inherit.aes = TRUE,
           ...) {
    ggplot2::layer(
      data = data,
      mapping = mapping,
      stat = stat,
      geom = GeomFlatViolin,
      position = position,
      show.legend = show.legend,
      inherit.aes = inherit.aes,
      params = list(trim = trim,
                    scale = scale,
                    ...)
    )
  }

#' @rdname ggplot2-ggproto
#' @format NULL
#' @usage NULL
#' @export
GeomFlatViolin <-
  ggproto(
    "GeomFlatViolin",
    Geom,
    setup_data = function(data, params) {
      data$width <- data$width %||%
        params$width %||% (resolution(data$x, FALSE) * 0.9)

      # ymin, ymax, xmin, and xmax define the bounding rectangle for each group
      data %>%
        dplyr::group_by(.data = ., group) %>%
        dplyr::mutate(
          .data = .,
          ymin = min(y),
          ymax = max(y),
          xmin = x,
          xmax = x + width / 2
        )
    },

    draw_group = function(data, panel_scales, coord)
    {
      # Find the points for the line to go all the way around
      data <- base::transform(data,
                              xminv = x,
                              xmaxv = x + violinwidth * (xmax - x))

      # Make sure it's sorted properly to draw the outline
      newdata <-
        base::rbind(
          dplyr::arrange(.data = base::transform(data, x = xminv), y),
          dplyr::arrange(.data = base::transform(data, x = xmaxv), -y)
        )

      # Close the polygon: set first and last point the same
      # Needed for coord_polar and such
      newdata <- rbind(newdata, newdata[1,])

      ggplot2:::ggname("geom_flat_violin",
                       GeomPolygon$draw_panel(newdata, panel_scales, coord))
    },

    draw_key = draw_key_polygon,

    default_aes = ggplot2::aes(
      weight = 1,
      colour = "grey20",
      fill = fColor,
      size = 0.5,
      alpha = NA,
      linetype = "solid"
    ),

    required_aes = c("x", "y")
  )
#######################
stopAndReturnError <- function(message, ...) {
	cat(message, file=paste0(outs, "/Error_Message_from_R.txt"))
	stop(message, ...)
}
#######################

#######################
robustLog2 <- function(x) {
	x <- as.numeric(x)
	if (x >= 0) {
		log(x + 1, 2)
	} else {
		0
	}
}
#######################
makeItRain <- function(p) {
  if(parViolin && parJitter && parBox){
      print("All the graphix!")
      p <- p + violinPart + jitterPart + boxPart +
      scaleShade + scaleFill + xlabel + ylabel + guide
      }else if(parViolin && parBox && !parJitter){
        print("Weather calls for boxy clouds")
        p <- p + violinPart + boxPart +
        scaleShade + scaleFill + guide + xlabel + ylabel
      }else if(parViolin && parJitter && !parBox){
         print("Jittery Violins - aka Staccato")
         p <- p + violinPart + jitterPart +
         scaleShade + scaleFill + guide + xlabel + ylabel
      }else if(parViolin && !parBox && !parJitter){
         print("Violin Solo")
         p <- p + violinSoloPart +
         scaleShade + scaleFill + guide + xlabel + ylabel
      }else if(parJitter && parBox && !parViolin) {
        print("Box and Whisker with a bit of Jitter")
        p <- p + jitterPart + boxPart +
        scaleShade + scaleFill + guide + xlabel + ylabel
      }else if(parJitter && !parViolin && !parBox){
        print("Jitter Only?! Go Home.")
        p <- p + jitterPart +
        scaleShade + scaleFill + guide + xlabel + ylabel
      }else if(parBox && !parJitter && !parViolin){
        print("Box and Whisker Only")
        p <- p + boxPart +
        scaleShade + scaleFill + guide + xlabel + ylabel
      }else{
        print("NO GRAPH OPTIONS SELECTED")
        p <- p +
        scaleShade + scaleFill + guide + xlabel + ylabel
  }
  return (p)
}
#######################
# Setting Default Variables for 'Nicer' Plots #
###############################################
rainGap <- 0.1
rainDropSize <- 2.5
jitterWidth <- 0.05
jitterHeight <-1.2
boxWidth <- 0.2
boxNudgeX <- -0.12
boxNudgeY <- 0
boxDodgeW <- 0.2
violinWidth <- 1.2

## Swapping Custom Variables Where Needed ##
if(parViolin && parJitter && parBox){
    print("All the graphix!")
    jitterHeight <- 1.2
    rainDropSize <- 0.5
    rainGap <- 0.15
    boxWidth <- 0.23
    violinWidth <- 1.1
    if(noKatPar){
        jitterHeight <- 1.2
        rainDropSize <- 0.5
        rainGap <- 0.09
        boxWidth <- 0.11
        violinWidth <- 1.1
        }else{
          boxWidth <- 0.17
          rainGap <- 0.09
          jitterHeight <- 1.4
    }
}else if(parViolin && parBox && !parJitter){
    print("Weather calls for boxy clouds")
    if(noKatPar) {
        boxWidth <- 0.12
        boxNudgeX <- 0.075
        }else{
          boxDodgeW <- 0.3
          boxWidth <- 0.2
          violinWidth <- 1.3
          boxNudgeX <- -0.19
    }
}else if(parViolin && parJitter && !parBox){
    print("Jittery Violins - aka Staccato")
    jitterWidth <- 0.05
    rainGap <- 0.09
    rainDropSize <- 0.5
    if(noKatPar) {
      rainGap <- 0.09
        }else{
            jitterWidth <- 0.09
            rainGap <- 0.01
            rainDropSize <- 0.4
            violinWidth <- 1.3
        }
}else if(parViolin && !parBox && !parJitter){
    print("Violin Solo")
    if (noKatPar) {
        }else{
    }
}else if(parJitter && parBox && !parViolin) {
    print("Box and Whisker with a bit of Jitter")
    rainDropSize <- 0.5
    jitterWidth <- 0.1
    boxWidth <- 0.25
    boxNudgeX <- 0.1

    if(noKatPar) {
         }else{
           boxWidth <- 0.4
           boxDodgeW <- 0.2
           jitterWidth <- 0.2
    }
}else if(parJitter && !parBox && !parViolin){
    print("Jitter Only?! Go Home.")
    jitterWidth <- 0.4
    rainDropSize <- 2.75
    if (noKatPar) {
         }else{
    }
}else if(parBox && !parJitter && !parViolin){
    print("Box and Whisker Only")
    boxWidth <- 0.5
    boxNudgeX <- 0
    boxDodgeW <- 0.5
    if (noKatPar) {
         }else{
    }
}else{
    print("NO GRAPH OPTIONS SELECTED")
}


##       GENERATE PLOTS      ##
###############################
input00 <- read.csv(inCSVPath, sep = ",", header = TRUE, check.names = FALSE)
if(!nrow(input00) > 0){
  print("Unable to read the input file correctly, try again.")
  stopAndReturnError("Unable to read the input file correctly.", call. = FALSE)
}
if ("EventNumberDP" %in% colnames(input00)){
  print("EventNumber Detected - This is FlowJo")
  input00 <- input00 %>%
    select(EventNumberDP, everything())
    colnames(input00) <- gsub( "(^.*)( :: )" , "", colnames(input00), ignore.case = TRUE )
    ImSeqGeeZee <- F
  } else {
     print("We're in SeqGeq")
     ImSeqGeeZee <- T
}



## UNIVERSAL MISC:

subornosub <- list()

if (!noKatPar) {
###############################################
## Test implement find popName
###############################################
packages(xml2)

## TODO - Make sure we're finding the XML in a robust way for SeqGeq (see ACS protocol below for clues). IT
#   Note: After prelim' investigation, the issue appears to be relate to the XML at hand, not the location...

var1 <- sub("[^/]+$", "", outPath)
var1
# var1 <- sub("[^/]+$", "", xmlPath)
var2 <- sub("(.*)/.*", "\\1", var1)
var2
var3 <- sub("[^/]+$", "", var2)
var3
var4 <- sub("(.*)/.*", "\\1", var3)
var4
if (ImSeqGeeZee){
    var4a <- sub("(.*-)","",var4)
    var4b <- sub(".geqzip", ".geq", var4a)
    var4c <- paste0(var3, var4b)
    var5 <- var4c
}else{
    var5 <- paste0(var4, ".wsp")
}


findingWorkspace <- F

# Check if the workspace exists - If not, assume we're in an ACS file:
if (file.exists(var5)) {
workspace <- read_xml(var5)

findingWorkspace <- T

} else { # (we're in an ACS file)

  var1 <- sub("[^/]+$", "", xmlPath)

    if(file.exists(var1)){

  if (ImSeqGeeZee){
    writeLines("In SeqGeq - GeqZip flavor = ACS")

      setwd(var1)
      files <- as.matrix(list.files(pattern = "\\.geq$"))
      colnames(files) <- "wsps"
      lastACSwsp <- files[nrow(files)]
      var2 <- paste0(var1,lastACSwsp)

  }else{ # you're in FlowJo... specifically an ACS file

    setwd(var1)
    files <- as.matrix(list.files(pattern = "\\.wsp$"))
    colnames(files) <- "wsps"
    lastACSwsp <- files[nrow(files)]
    var2 <- paste0(var1,lastACSwsp)

  }

  if (file.exists(var2) && (nrow(files) > 0) ) {
      workspace <- read_xml(var2)
      findingWorkspace <- T
  }
    }
}


if(findingWorkspace) {
getSamList <- function(obj) {
    xml_child(obj, "SampleList")
}

# get all samples in the workspace
samList <- getSamList(workspace)

# fix file names that contain space characters:
selFile00 <- sub(" ", "%20", selFile)

# get subset of samples; the sample of interest
uri <- paste0("..//Sample/DataSet[contains(@uri,'", selFile00, "')]") # use the new var selFile from Java to find the sample's ID
uriNode <- xml_find_first(samList, uri)
tbl <- ldply(xml_attrs(uriNode))
SampleID <- tbl$V1[2]

# Use the sampleID to find the subpopulation selected:


popName <- paste0(".//SampleNode[@sampleID = ", "'", SampleID, "']")
samNode <- xml_find_first(samList, popName)


# Note: 'sampleName' is actually the name of the population (a sample OR subset) exported as part of of the plugin process
populationName <- sub(" ", "%20", sampleName) ## IMPORTANT NOTE: We should not be using this as a population name (use nodeElement instead)

nodeElement <- sub("(.* of )", "", extNodeEl)

# get the name of the subpopulation selected
selFile00 <- sub("\\.fcs","",selFile00)
tester <- sub(paste0(selFile00,"\\."), "", nodeElement)
popFinder <- paste0(".//node()[@name = '",tester,"']")
xorsize <- xml_find_all(samNode, popFinder)

# get the ID of the subpopulation selected
parentNode <- xml_find_first(xorsize, ".//Population/Gate")

if(length(parentNode) > 1){

    parentNodeIDs <- ldply(xml_attrs(parentNode[[2]]))

    if(anyNA(parentNodeIDs)) {
      parentNodeIDs <- ldply(xml_attrs(parentNode[[1]]))
    }

    parentNodeID <- parentNodeIDs$V1[[1]]
    if(length(parentNodeIDs$V1) > 1){
        parentNodeIDalt <- parentNodeIDs$V1[[2]]
    } else {
        parentNodeIDalt <- "NA"
    }

}else{
parentNodeIDs <- ldply(xml_attrs(parentNode))
parentNodeID <- parentNodeIDs$id
parentNodeIDalt <- parentNodeIDs$parent_id
}


if (is.null(parentNodeIDalt)) {
  parentNodeIDalt <- "NA"
}

# find out if that node has any (sub) populations:
subfinder <- paste0("./Subpopulations")
subornosub <- xml_find_all(xorsize, subfinder)

if (length(subornosub) > 0) {

  writeLines("your selected population has subsets, worth investigating!")

  findClustyPops <- paste0(".//Population")
  crustyPops <- xml_find_all(subornosub, findClustyPops)

    names <- ldply(xml_attrs(crustyPops))["name"]
    numbers <- ldply(xml_attrs(crustyPops))["count"]


    # let's id parents and child gate IDs for checks in the next section
    popNameDF <- ldply(xml_attrs(crustyPops, "name"))

    gateName <- xml_find_all(crustyPops, "./Gate[node()]")

    pGateIDs <- ldply(xml_attr(gateName, "parent_id"))
    colnames(pGateIDs) <- "parentID"

    gateIDs <- ldply(xml_attr(gateName, "id"))
    colnames(gateIDs) <- "ID"

    nameTable <- cbind(names,numbers,pGateIDs,gateIDs)

    # let's check that the these populations are drawn on the catPar dimension
    findPopDim01 <- paste0(".//Gate")

    popDims02 <- xml_find_all(crustyPops, findPopDim01)

    ## Need to find another way of identifying the id attributes here:
    # attrsX = ldply(xml_attrs(popDims02))["id"]

        # This is looking good!
        attrsX <- ldply(xml_attr(popDims02, "id"))
        attrsX2 <- ldply(xml_attr(popDims02, "parent_id"))
        colnames(attrsX) <- "id"

        dimsY = as.data.frame(unlist(xml_attrs(xml_find_first(xml_find_all(crustyPops,
                                                                           ".//*/gating:RectangleGate"),
                                                              ".//*/data-type:fcs-dimension"))))
        colnames(dimsY) = "gDims"

        rossetaDF <- cbind(attrsX, dimsY)

        bsFilter <- subset(rossetaDF, gDims==catPar)["id"]

        pr <- data.frame(name = character(),
                         count = character(),
                         parentID = character(),
                         stringsAsFactors = F)

    if(nrow(bsFilter) > 0){
        po <- 0
        for (og in 1:nrow(nameTable)) {
            for (go in 1:nrow(bsFilter)){
                if ((bsFilter$id[go] == nameTable$ID[og]) && !is.na(nameTable$ID[og])){
                    po <- po+1
                    pr[po,] <- data.frame(name = nameTable$name[og],
                    count = nameTable$count[og],
                    parentID = nameTable$parentID[og],
                    stringsAsFactors = F)
                }
            }
        }
    }

        # Let's check that the number of rows isnt the same as the number of populations we're looking for (if so, the we may not need to do so much checking):

        rowNum <- nrow(pr)

        popNum <- length(unique(input00[[catPar]]))

        if (rowNum == popNum) {

          qr <- data.frame(name = character(),
                           count = character(),
                           stringsAsFactors = F)

          op <- 0
          for (z in 1:nrow(pr)) {
            if (pr$count[z] > 0){
              op <- op+1
              qr[op,] <- data.frame(name = pr$name[z],
                                    count = pr$count[z],
                                    stringsAsFactors = F)
            }
          }

        } else {

            # Let's check that the parent of these populations is the same population id as the selected population
            lParent <- as.character(parentNodeID)

            qr <- data.frame(name = character(),
                             count = character(),
                             stringsAsFactors = F)

            op <- 0
            for (z in 1:nrow(pr)) {
              if (lParent == pr$parentID[z] && !is.na(pr$parentID[z])){
                op <- op+1
                qr[op,] <- data.frame(name = pr$name[z],
                                      count = pr$count[z],
                                      stringsAsFactors = F)
              }
            }

            if (nrow(qr)==0){
              mParent <- as.character(parentNodeIDalt)
              op <- 0
              for (zz in 1:nrow(pr)) {
                if (mParent == pr$parentID[zz] && !is.na(pr$parentID[zz])){
                  op <- op+1
                  qr[op,] <- data.frame(name = pr$name[zz],
                                        count = pr$count[zz],
                                        stringsAsFactors = F)
                }
              }
            }

            ## Deal with NA values in parentID column if the selected parent is a Boolean population
            if (nrow(pr) > 0){
                if(nrow(qr)==0){

                    for (olo in 1:nrow(pr)) {
                        if (is.na(pr$parentID[olo])) {
                            pr$parentID[olo] <- "NA"
                        }
                    }

                    op <- 0
                    for (zz in 1:nrow(pr)) {
                        if (mParent == pr$parentID[zz] && !is.na(pr$parentID[zz])){
                            op <- op+1
                            qr[op,] <- data.frame(name = pr$name[zz],
                            count = pr$count[zz],
                            stringsAsFactors = F)
                        }
                    }
                }
            }
        }
          dupCount <- T
          ## Deal with non-unique counts:
          if (nrow(qr) > 1) {
            if (any(duplicated( qr$count ))) {
              dupCount <- T
            } else {
              dupCount <- F
            }
          }
       }
    } else {
      writeLines("We couldn't find your workspace file go on with creating the ViolinBox plots without naming subpopulations nicely.")
    }
}




## Check if all the population counts are zero and fail gracefully if so. This check is only going to work if we also find the population names.
zeroPops <- F
if (length(subornosub) > 0) {
  if (nrow(qr) > 0){
    # if (dupCount == F) {
      zeroPops <- T
      for (v in 1:nrow(qr)) {
        if(qr$count[v] == 0){
        } else {
          zeroPops <- F
        }
      }
      if (zeroPops) {
        stopAndReturnError("Detecting zero cells for all populations in categorical parameter.\n\n Fix current clusters or try reclustering and re-run this plugin.\n\n")
      }
    # }
  }
}

if (ImSeqGeeZee == F) {
  input00$EventNumberDP <- NULL
}

if (!noKatPar) {
    cnam <- colnames(input00)
    cnam01 <- sub(catPar, NA, cnam)
    ## Check if the user selected any parameters!
    if (!length(input00) > 1) {
        stopAndReturnError("Forget something?...\nNo parameters for plotting detected in R.\n\nTry deleting and then restarting the plugin,\nmaking sure to select some parameters in\nthe top parameter selector dialog there.\n\nCheers!")
    }
    if (groupByClust){                                                 ## Thanks Dylan! :)
        numUniqueParams <- ncol(input00)-1
    } else{
        numUniqueParams <- length(row_number(unique(input00[catParam])))
    }
}else{
    ## Check if the user selected any parameters!
    if (!length(input00) > 0) {
        stopAndReturnError("Forget something?...\nNo parameters for plotting detected in R.\n\nTry deleting and then restarting the plugin,\nmaking sure to select some parameters in\nthe top parameter selector dialog there.\n\nCheers!")
    }
    numUniqueParams <- length(unique(col(input00)))
}

# Color Blind Pallette Idea:
# library(rcartocolor)
# nColor <- 12
# scales::show_col(carto_pal(nColor, "Safe"))
# scaleShade <- scale_colour_brewer(palette = "Set2")
# scaleFill <-  scale_fill_brewer(palette = "Set2")

## This color palette generator requires a minimum of 4 parameters, so:
  if (numUniqueParams < 4) {
    numUniqueParams <- 4
  }

getPalette <- colorRampPalette(brewer.pal(8, "Dark2"))(numUniqueParams)

    xLabel <- "Parameters"

        if (groupByClust) {
          xLabel <- paste0(catPar, " Clusters")
           theme02 <- theme(axis.text.x = element_text(angle = 48, hjust = 1.1, vjust = 1, size = 15, colour = jColor),
                            axis.text.y = element_text(size = 15, colour = jColor),
                            strip.text = element_text(size = 20),
                            legend.text = element_text(size = 15, colour = jColor),
                            legend.key = element_rect(fill = fColor, colour = NULL))
        }

scaleShade <- scale_colour_manual(values = getPalette)
scaleFill <-  scale_fill_manual(values = getPalette)
ylabel <-	ylab('Log2Expression')
xlabel <-  xlab(xLabel)
if (parFlipGraph) {
 theme01 <- theme(axis.text.x = element_text(angle = 0, vjust = 0.5, size = 15, colour = jColor),
                  axis.text.y = element_text(size = 15, colour = jColor),
                  strip.text = element_text(size = 20),
                  legend.text = element_text(size = 15, colour = jColor),
                  legend.key = element_rect(fill = fColor, colour = NULL))
}else{
  theme01 <- theme(axis.text.x = element_text(angle = 90, vjust = 0.5, size = 15, colour = jColor),
                   axis.text.y = element_text(size = 15, colour = jColor),
                   strip.text = element_text(size = 20),
                   legend.text = element_text(size = 15, colour = jColor),
                   legend.key = element_rect(fill = fColor, colour = NULL))
}

if (noKatPar){

  input <- input00
  print("No Categorical Identified")

  # Use gather function to quickly create the data frame.
  violinPlotDF <- gather(input00, key = "Gene", value = "Log2Expression", 1:ncol(input00))
  if(ncol(violinPlotDF) < 2){
    print("Unable to read the inputs correctly, check parameter names, or that there is a clustering parameter.")
    stopAndReturnError("Unable to read the inputs correctly, check parameter names.", call. = FALSE)
  }
  # can use sapply to apply the robustlog2 to Values
  tempDF <- sapply(violinPlotDF$Log2Expression, robustLog2)
  tempDF <- data.frame(matrix(unlist(tempDF), nrow=length(tempDF), byrow=T))
  violinPlotDF <- cbind(violinPlotDF, tempDF)
  violinPlotDF <-  violinPlotDF[,-2]
  colnames(violinPlotDF)[2] <- "Log2Expression"

  violinPlotDF$Log2Expression <- as.numeric(violinPlotDF$Log2Expression)
  violinPlotDF$Gene <- as.factor(violinPlotDF$Gene)

p <- ggplot(violinPlotDF,aes(x=Gene,y=Log2Expression, fill = Gene, colour = Gene))
    violinSoloPart <- geom_violin(position = position_nudge(x = .1, y = 0),
                                  adjust = 2.5,
                                  trim = FALSE,
                                  width = 1.2)

    violinPart <- geom_flat_violin(position = position_nudge(x = rainGap, y = 0),
                                   adjust = 2.5,
                                   trim = FALSE,
                                   width = violinWidth)

    if (parJitter && parBox && !parViolin) {
          jitterPart <- geom_jitter(aes(x = Gene,y = Log2Expression),
          position = position_jitter(width = jitterWidth),
          alpha = 0.1,
          colour = jColor)

        }else{
          jitterPart <- geom_point(alpha = .10,
          position = position_jitter(width = jitterWidth,
          height = jitterHeight),
          size = rainDropSize)
    }

    if (parJitter && parBox && !parViolin) {
         boxPart <- geom_boxplot(aes(x = Gene,
         y = Log2Expression),
         outlier.colour = NA,
         width = boxWidth,
         position = position_dodge(preserve = "total",width = boxDodgeW),
         colour = jColor)

       }else{
         boxPart <- geom_boxplot(aes(x = Gene,
         y = Log2Expression),
         position = position_nudge(x = boxNudgeX,y = boxNudgeY),
         outlier.shape = NA,
         alpha = 0.3,
         width = boxWidth,
         colour = jColor)
    }

    xLabel <- "Parameters"

        if (groupByClust) {
          xLabel <- paste0(catPar, " Clusters")
          theme01 <- theme02
        }

# misc:
ylabel <-	ylab('Log2Expression')
xlabel <-  xlab(xLabel)
theme <- theme_cowplot()
guide <- guides(fill = FALSE, colour = FALSE)
p <- makeItRain(p)
p <- p + theme + theme01
 if (parDarkTheme) {
 p <- p + theme_dark() + theme(axis.text.x = element_text(size = 15, colour = jColor),
                               axis.text.y = element_text(size = 15, colour = jColor),
                               strip.text = element_text(size = 20),
                               plot.background = element_rect(fill = fColor),
                               legend.background = element_rect(fill = fColor),
                               panel.background = element_rect(fill = fColor),
                               legend.key = element_rect(fill = fColor, colour = NULL))
 }

} else if (!zeroPops) {
    # TODO: b/c of possible long categorical param names moving the legend position, need to shorten catParam names.
     #TODO (continued)# or better yet! - Create a legend figure that's separate from the image, and moveable... IT

# Lets take the selected Modal Parameter column and move it to the front of the DF before running gather()
input0 <- input00 %>% dplyr::select(catParam, everything())

# Get rid of pesky clusters at Zero (those should be cells not included in clustering which snuck in here)
mainCatNotZero <- subset(input0, input0[[catPar]]!=0)

input <- mainCatNotZero

# Make sure that we generated the necessary output from XML parsing
if (length(subornosub) > 0) {
if (nrow(qr) > 0){
if (dupCount == F) {

## TODO - Make sure proper checks are in place to prevent incorrect associatsions. IT
# Swap out correct names from catParam populations with catParam numbers:

library(plyr)
DFTable <- plyr::count(input[[catPar]])

uCat <- unique(input[[catPar]])

q <- data.frame( newName = character(),
                 oldName = character(),
                 count = character(),
                 stringsAsFactors = F )

for (r in 1:length(uCat)) {

   for (l in 1:nrow(qr)) {
     if (qr$count[l] == DFTable$freq[r]) {
     q[l,] <- data.frame(newName = qr$name[l],
                         oldName = DFTable$x[r],
                         count = qr$count[l],
                         stringsAsFactors = F )
     }
   }
 }


# TODO - Think about adapting this to non-contiguous populations
# Take into account populations in the parent not defined in the subpopulations -
#     shouldn't need to worry about zeros as these have already been removed
nq <- nrow(q)
for (t in 1:length(uCat)) {
  if (is.na(q$oldName[DFTable$x[t]])){
    nq <- nq+1
    q[nq,] <- data.frame(newName = DFTable$x[t],
                         oldName = DFTable$x[t],
                         count = DFTable$freq[t],
                         stringsAsFactors = F)
  }
}


            # Now make the appropriate swap (exact match regex):
            for (i in 1:length(uCat)) {
              input[[catPar]] <- gsub(paste0("^",q$oldName[i],"$"), q$newName[i], input[[catPar]])
            }


            # Check if the q table has any NA values and if so, abort the renaming of populations function, too many unknowns. IT
            # TODO - Make sure we force the researcher to save their workspace before running this plugin, or get the xml from a more reliable source (above).

            any(is.na(q))

            if (any(is.na(q))){
              input <- input0
            }
        }
    }
} else {
    writeLines("Categorical Populations are empty (zero counts).")
}

# Use gather function to quickly create the data frame.
violinPlotDF <- gather(input, key = "Gene", value = "Log2Expression", 2:ncol(input))
if(ncol(violinPlotDF) < 3){
  print("Unable to read the inputs correctly, check parameter names, or that there is a clustering parameter.")
  stopAndReturnError("Unable to read the inputs correctly, check parameter names.", call. = FALSE)
}

# can use sapply to apply the robustlog2 to Values
tempDF <- sapply(violinPlotDF$Log2Expression, robustLog2)
tempDF <- data.frame(matrix(unlist(tempDF), nrow=length(tempDF), byrow=T))
violinPlotDF <- cbind(violinPlotDF, tempDF)
violinPlotDF <-  violinPlotDF[,-3]
colnames(violinPlotDF)[3] <- "Log2Expression"
violinPlotDF$Log2Expression <- as.numeric(violinPlotDF$Log2Expression)
violinPlotDF$Gene <- as.factor(violinPlotDF$Gene)
violinPlotDF[catParam] <- as.factor(gsub("\\s*", "", as.matrix(violinPlotDF[catParam])))
rm(tempDF)

ttle <- catPar

if (groupByClust) {
  violinPlotDF.saveGame <- violinPlotDF
    colnames(violinPlotDF) <- c("Gene", as.character(catPar), "Log2Expression")
  ttle <- "Parameter"
  theme01 <- theme02
}

  # The plot:
  p <- ggplot(violinPlotDF, aes(x = Gene, y = Log2Expression), text) # , fill = violinPlotDF[,1]))
  #optional aspects of the plots:
  violinSoloPart<- geom_violin(aes(fill = violinPlotDF[[catPar]]),
                               position = position_nudge(x = .2,y = 0),
                               adjust = 2,
                               trim = FALSE,
                               alpha = .5,
                               colour = NA,
                               width = 1.2)

  violinPart <- geom_flat_violin(aes(fill = violinPlotDF[[catPar]]),
                                 position = position_nudge(x = rainGap, y = 0),
                                 adjust = 2,
                                 trim = FALSE,
                                 alpha = .5,
                                 colour = NA,
                                 width = violinWidth)

  if(parJitter && parBox && !parViolin) {
     jitterPart <- geom_jitter(aes(x = Gene, y = Log2Expression,
                              fill = violinPlotDF[[catPar]]),
                              position = position_jitter(width = jitterWidth),
                              size = rainDropSize,
                              alpha = 0.1,
                              colour = jColor,
                              size = 0.1)

     } else if (parJitter && !parBox && !parViolin){

          jitterPart <- geom_point(aes(x = Gene,y = Log2Expression,
                                         fill = violinPlotDF[[catPar]]),
                                         alpha = .1,
                                         position = position_jitter(width = jitterWidth, height = jitterHeight),
                                         size = rainDropSize,
                                         shape = 21,
                                         alpha = 0.1,
                                         colour = jColor,
                                         size = 0.1)

     }else{
      jitterPart <- geom_point(aes(x = as.numeric(Gene)-.15,y = Log2Expression,
                               fill = violinPlotDF[[catPar]]),
                               alpha = .1,
                               position = position_jitter(width = jitterWidth, height = jitterHeight),
                               size = rainDropSize,
                               shape = 21,
                               alpha = 0.1,
                               colour = jColor,
                               size = 0.1)
  }
  if(parJitter && parBox && !parViolin) {
    boxPart <- geom_boxplot(aes(x = Gene,y = Log2Expression,
                            fill = violinPlotDF[[catPar]]),
                            outlier.colour = NA,
                            width = boxWidth,
                            position = position_dodge(preserve = "total",width = boxDodgeW),
                            colour = jColor,
                            size = 0.1)

    }else{
      boxPart <- geom_boxplot(aes(x = Gene,y = Log2Expression,fill = violinPlotDF[[catPar]]),
                            position = position_dodge(preserve = "total",width = boxDodgeW),
                            outlier.shape = NA,
                            width = boxWidth,
                            colour = jColor,
                            size = 0.1)
  }
  ##misc:

    guide <-  guides(fill=guide_legend(title=ttle)) #TODO shorten the length of legend title
    theme <- theme_cowplot()
    p <- makeItRain(p) + theme + theme01

}


###  TODO - Add Stat Comparisons:
##     * Remove non-statistically significant comparisons
#################################
## Add Statistical Comparisons ##
#################################

# groupByClust <- F
# statCompare <- T
if (statCompare) {

  packages(gginnards)
  packages(tidyverse)
  packages(ggsignif)
  packages(broom)

  maximus <- max(violinPlotDF$Log2Expression)
  minimus <- min(violinPlotDF$Log2Expression)

  factor <- "Gene"

  groups <- as.character(unique(violinPlotDF[[factor]]))

  expand.grid.unique <- function(x, y, include.equals=FALSE){
    x <- unique(x)
    y <- unique(y)
    g <- function(i){
      z <- setdiff(y, x[seq_len(i-include.equals)])
      if(length(z)) cbind(x[i], z, deparse.level=0)
    }
    do.call(rbind, lapply(seq_along(x), g))
  }

  combs <- as.data.frame(expand.grid.unique(groups, groups), stringsAsFactors=FALSE)

  my.comps <- as.data.frame(t(combs), stringsAsFactors=FALSE)
  colnames(my.comps) <- NULL
  rownames(my.comps) <- NULL

  my_comparisons <- as.list(my.comps)
  # my_comparisons <- list(catPar)

  height.y = c((maximus+0.05*maximus), (maximus+0.11*maximus), (maximus+0.17*maximus))

  symn <- list(cutpoints = c(0, 0.0001, 0.001, 0.01), symbols = c("****", "***", "**", "*"))
  # symn <- list(cutpoints = c(0, 0.0001, 0.001, 0.01, 0.05), symbols = c("****", "***", "**", "*"))

  if (parOverlayPlots) {
      p <- p + stat_compare_means(comparisons = my_comparisons,
                                  size=3,
                                  label = "p.signif",
                                  hide.ns = TRUE,
                                  symnum.args <- symn,
                                  na.rm = TRUE,
                                  step.increase = 0.2,
                                  show.legend = TRUE,
                                  color = jColor)
  }

  if (!noKatPar) {

      ## Adjust the position of the KW value
        if (parOverlayPlots) {
          ykw <- minimus-0.23*maximus

        } else {
          ykw <- minimus+0.05*maximus
        }

      ## only run KW test if there's a categorical parameter:
                    p <- p + stat_compare_means(label.x = 1.4,
                                                 label.y = (ykw),
                                                 size=2.3,
                                                 colour = jColor)

  ## With a categorical parameter the y-axis max spacing needs some help:
   p <- p + scale_y_continuous(expand = expand_scale(mult = c(.1, .2)))

    }
  }
#################################
#################################


if(parOverlayPlots && !noKatPar){
  p <- p + facet_wrap(~violinPlotDF[[catPar]])
}

if (parFlipGraph) {
	p <- p + coord_flip()
}

if(parDarkTheme && !noKatPar){
  p <- p + theme_dark() + theme01 + theme(strip.text = element_text(size = 18),
                                          plot.background = element_rect(fill = fColor),
                                          legend.background = element_rect(fill = fColor),
                                          panel.background = element_rect(fill = fColor),
                                          legend.key = element_rect(fill = fColor, colour = NULL))
}

if ( parJitter || parViolin || parBox) {
    ggsave(paste0(outPath, ".png"),  plot = p, width = 500, height = 270, units = "mm", dpi = 320, device = "png")
}

if(parCreatePDF) {
    ggsave(paste0(outPath, ".pdf"),  plot = p, width = 500, height = 270, units = "mm", device = "pdf")
}

#######################
### Heatmap options ###
pdf(NULL) # must have this for windows machine to make plotting devices available


if(parHeatmap || parScaledHM) {
  options(bitmapType = 'cairo', device = 'png') # needed for Windows when using ggsave with multiple images

  if (parHeatmap){
    scl <- "none"
  } else {    ## parScaledHM
    scl <- "column"
  }

  if(!noKatPar) {
    groupedInput <- input %>%
      group_by_at(vars(catParam)) %>%
      summarise_all(mean)

    namers <- groupedInput[,catParam]
    namerss <- (apply(namers,2,function(x)gsub(' ', '',x)))

    workDoneFlag <- F
    if (length(subornosub) > 0) {
      if (nrow(qr) > 0){
        writeLines("you've got the populations of interest")
      } else {
        namerss <- t(as.data.frame(lapply(namerss,function(x)gsub(x, paste0(catParam," ", x), x))))
        workDoneFlag <- T
      }
    } else {
      if (!workDoneFlag) {
        namerss <- t(as.data.frame(lapply(namerss,function(x)gsub(x, paste0(catParam," ", x), x))))
      }
    }

    rownames(groupedInput) <- namerss
    groupedInput[catParam] <- NULL

    GI <- groupedInput

    # indexList <- which(!names(groupedInput) %in% c(catParam))
    indexList <- which(!names(GI) %in% c(catParam))

    if(ImSeqGeeZee){
      print("Unscaled Heatmap with Categorical Parameter for SeqGeq")
      hm <- pheatmap::pheatmap(GI[,indexList],
                               color = inferno(n=512, begin = 0, end = 0.9),
                               border_color = "grey20",
                               main = "",
                               show_rownames = TRUE,
                               show_colnames = TRUE,
                               fontsize_col = 12,
                               angle_col = "45",
                               fontsize_row = 12,
                               labels_row = namerss,
                               scale = scl)
    }else{
      print("Unscaled Heatmap with Categorical Parameter for FlowJo")
      hm <- pheatmap::pheatmap(GI[,indexList],
                               color = inferno(n=512, begin = 0, end = 0.9),
                               border_color = "grey20",
                               main = "",
                               show_rownames = TRUE,
                               show_colnames = TRUE,
                               fontsize_col = 12,
                               angle_col = "45",
                               fontsize_row = 12,
                               labels_row = namerss,
                               legend_labels = "checkit",
                               scale = scl)
    }
  }else if(noKatPar) {
    packages(matrixStats) #Then it does NOT have categorical parameter, so massage data in another way...
    if(ImSeqGeeZee == F){
      input2 <- as.matrix(input)
    }else{
      input2 <- as.matrix(input)
    }

    if(ImSeqGeeZee){
      print("Unscaled Heatmap without Categorical Parameterfor SeqGeq")
      hm <- pheatmap::pheatmap(input2,color = inferno(n=512, begin = 0, end = 0.9),
                               border_color = "grey20",
                               main = "",
                               show_rownames = FALSE,
                               show_colnames = TRUE,
                               fontsize_col = 8,
                               angle_col = "45",
                               fontsize_row = 8,
                               scale = scl,
                               kmeans_k = 50)
    }else{
      print("Unscaled Heatmap without Categorical Parameter for FlowJo")
      hm <- pheatmap::pheatmap(input2,color = inferno(n=512, begin = 0, end = 0.9),
                               border_color = "grey20",
                               main = "",
                               show_rownames = FALSE,
                               show_colnames = TRUE,
                               fontsize_col = 8,
                               angle_col = "45",
                               fontsize_row = 8,
                               scale = scl,
                               kmeans_k = 50)
    }
  }

    scale <- 47

    dHM <- dim(GI)
    nCat <- dHM[1]
    nPar <- dHM[2]

    if(nCat > nPar){
        hScale <- scale
        wScale <- (nPar/nCat)*scale + 1
    } else if(nPar > nCat){
        hScale <- (nCat/nPar)*scale
        wScale <- scale + 1
    } else {
        hScale <- scale
        wScale <- scale + 1
    }

    ggsave(paste0(outPath, ".HM.png"),  plot = hm, width = wScale, height = hScale, device = "png")
}

## Create Stats Table ##
########################
## Goals:
#   Mean, Median, SD, Min, Max, Count, and Frequency
if (createStatTable){
  packages(magrittr)
  packages(qwraps2)
    packages(htmlTable)

  simpleSummaryTable <-
    list("Pop Size" =
           list("count" = ~ NROW(.data$Log2Expression),
                "frequency (%)" = ~ (NROW(.data$Log2Expression)/NROW(violinPlotDF))*100,
                "sum" = ~ sum(.data$Log2Expression)),
         "Range" =
           list("min" = ~ min(.data$Log2Expression),
                "max" = ~ max(.data$Log2Expression)),
         "Expression" =
           list("mean" = ~ mean(.data$Log2Expression),
                "median" = ~ median(.data$Log2Expression),
                "std-dev" = ~ sd(.data$Log2Expression))
    )

    if (noKatPar) {
      sTable <- summary_table(group_by(violinPlotDF, Gene), simpleSummaryTable)
      print (sTable)
      unique(violinPlotDF$Gene)
      colnames(sTable) <- c(unique(as.character(violinPlotDF$Gene)))
      # rtitle(sTable) = paste0(catParam, " ", n)

    } else {
      rm(nTable, nSpace, nHead)
      nTable <- list()
      sTable <- list()
      nTable <- simpleSummaryTable
      sTable <- NULL

      if(catParRaw != "[]" && catParRaw != "") {
          numUniqueParams <- length(row_number(unique(violinPlotDF[[catParam]])))
          serWT <- unique(violinPlotDF[[catParam]])
        }else{
          numUniqueParams <- length(unique(col(violinPlotDF)))
          serWT <- unique(col(violinPlotDF))
        }

      serWT <- data.frame(unique(violinPlotDF[[catParam]]))

      sWT <- data.frame( catPar = character(),
                         stringsAsFactors = F)

      for (k in 1:nrow(serWT)){
        sWT[k,] <- data.frame( catPar = gsub(" ","", serWT[k,]),
                               stringsAsFactors = F)
      }

      for (tt in unlist(sWT)) {
        print(tt)
      }

      for (n in unlist(sWT)) {

        nTable <- summary_table(group_by(filter(violinPlotDF, violinPlotDF[catParam] == n), Gene), simpleSummaryTable)
        ## TODO: Customize Table Rows (etc). IT

        colnames(nTable) <- c(unique(as.character(violinPlotDF$Gene)))

        # nHead <- data.frame(paste0(catParam, " ", n))

        nHead <- data.frame()[1,length(violinPlotDF$Gene)]
        for (m in 1:length(unique(as.character(violinPlotDF$Gene)))) {
          nHead <- cbind(nHead, "%%%%%%")
        }
        colnames(nHead) <- c(unique(as.character(violinPlotDF$Gene)))

        workDoneFlag <- F
        if (length(subornosub) > 0) {
        if (nrow(qr) > 0){
            row.names(nHead) <- n
        } else {
            row.names(nHead) <- paste0(catParam, " ", n)
            workDoneFlag <- T
        }
        } else {
            if (!workDoneFlag) {
                row.names(nHead) <- paste0(catParam, " ", n)
            }
        }

        nSpace <- data.frame()[1,length(violinPlotDF$Gene)]
        for (l in 1:length(unique(as.character(violinPlotDF$Gene)))) {
          nSpace <- cbind(nSpace, " ")
        }
        colnames(nSpace) <- c(unique(as.character(violinPlotDF$Gene)))
        row.names(nSpace) <- "."
        nTable <- rbind(nHead, nTable)
        nTable <- rbind(nSpace, nTable)
        sTable <- rbind(sTable, nTable)
      }
    }

  sTable <- txtRound(sTable,2)

  # newTable <- qwraps2::qable(sTable, rtitle = "Summary Statistics")

  packages(htmltools)
  packages(kableExtra)
  if (noKatPar) {
    cap <- "Basic statistics table"
  } else {
    cap <- paste0("Basic statistics table of ", catParam, " populations")
  }

  options(knitr.table.format = "html")
  khtmlTab <- kable(sTable, caption = cap) %>%
    kable_styling(bootstrap_options = "striped", full_width = F, position = "center", fixed_thead = T)
    suppressMessages( save_kable(khtmlTab, file = paste0(outPath,".html"), self_contained = T) )
}

# Close out of the test for parameters up at line ~578:
# }

# save(violinPlotDF, file=paste0(outPath, ".viDF.RData"))
# R seems to be saving the .RData when exiting, so let's clean up to at least make that tiny (i.e., empty environment)
# rm(list=ls())